# -*- coding: utf-8 -*-
"""
Created on nov. 28, 2018, 14:39
Copyright François Durand 2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from scipy.spatial import distance
from svvamp.preferences.generator_profile import GeneratorProfile
from svvamp.preferences.profile import Profile


class GeneratorProfileGaussianWell(GeneratorProfile):
    """Profile generator using the 'Gaussian well' model.

    Parameters
    ----------
    n_v : int
        Number of voters.
    n_c : int
        Number of candidates.
    sigma : list or ndarray
        1d array of numbers. The variance of the gaussian distribution along each dimension.
    shift : list or ndarray
        1d array of numbers, same dimension as ``sigma``. Shift for the mean position of the candidates.
    sort_voters : bool
        This argument is passed to :class:`Profile`.

    Notes
    -----
    Let us note ``n_dim`` the number of elements in ``sigma``. For voter ``v`` (resp. each candidate ``c``) and each
    axis ``i`` in ``range(n_dim)``, a position ``x_i[v]`` (resp. ``y_i[c]``) is independently drawn according to a
    normal distribution of mean 0 and variance ``sigma[i]``. If ``shift`` is used, the distribution of positions for
    candidates is displaced by this vector.

    Let ``d[v, c]`` denote the Euclidean distance between voter ``v``'s position ``x[v]`` and candidate ``c``'s
    position ``y[c]``. Then ``preferences_ut[v, c] = A - d[v, c]``, where ``A`` is such that the average utility is 0
    over the whole population.

    Remark: if ``n_dim = 1``, the population is single-peaked.

    Examples
    --------
        >>> generator = GeneratorProfileGaussianWell(n_v=10, n_c=3, sigma=[1], shift=[10])
        >>> profile = generator()
        >>> profile.preferences_rk.shape
        (10, 3)
    """

    def __init__(self, n_v, n_c, sigma, shift=None, sort_voters=False):
        self.n_v = n_v
        self.n_c = n_c
        self.sigma = np.array(sigma)
        self.n_dim = len(self.sigma)
        self.shift = np.zeros(self.n_dim) if shift is None else np.array(shift)
        self.log_creation = ['Gaussian well', self.n_c, self.n_v, 'Sigma', self.sigma,
                             'Shift', self.shift, 'Number of dimensions', self.n_dim]
        super().__init__(sort_voters=sort_voters)

    def __call__(self):
        voters_positions = np.random.randn(self.n_v, self.n_dim) * self.sigma
        candidates_positions = self.shift + np.random.randn(self.n_c, self.n_dim) * self.sigma
        preferences_utilities = - distance.cdist(voters_positions, candidates_positions, 'euclidean')
        preferences_utilities -= np.average(preferences_utilities)
        return Profile(preferences_ut=preferences_utilities, log_creation=self.log_creation,
                       sort_voters=self.sort_voters)



class _Options:
    def __init__(self):
        self.ERROR_WHEN_UNCOVERED_CODE = False
        """bool. If True, throw an AssertionError when an uncovered portion of code is reached."""


OPTIONS = _Options()



# -*- coding: utf-8 -*-
"""
Created on 10 dec. 2018, 13:34
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleKimRoush(Rule):
    """Kim-Roush method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleKimRoush()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[-1. -2. -2.]
         [-5. inf inf]]
        candidates_by_scores_best_to_worst
        [0 1 2]
        scores_best_to_worst
        [[-1. -2. -2.]
         [-5. inf inf]]
        w = 0
        score_w = [-1. -5.]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (False) ==> maj_fav_c_rk (False)       ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (False) ==> majority_favorite_c_ut (False)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (False)                ==>                IgnMC_c (False)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (False)                ==>                InfMC_c (False)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = nan
        log_icm: icm_option = lazy
        candidates_icm =
        [ 0.  0. nan]
        necessary_coalition_size_icm =
        [0. 0. 0.]
        sufficient_coalition_size_icm =
        [ 0. inf inf]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = lazy, um_option = lazy, icm_option = lazy, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 0. 0.]
        sufficient_coalition_size_cm =
        [0. 2. 3.]

    Notes
    -----
    At each round, all candidates with a Veto score strictly lower than average are simultaneously eliminated. When
    all remaining candidates have the same Veto score, the candidate with lowest index is declared the winner.

    Kim-Roush method does not meets InfMC.

    * :meth:`is_cm_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_icm_`: Non-exact algorithm from superclass :class:`Rule`.
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Statistical Manipulability of Social Choice Functions', K.H. Kim and F.W. Roush, 1996.
    """

    full_name = 'Kim-Roush'
    abbreviation = 'KR'

    options_parameters = Rule.options_parameters.copy()

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            log_identity="KIM-ROUSH", **kwargs
        )

    @cached_property
    def _count_ballots_(self):
        """
        Case where all the candidates have the same score:

            >>> profile = Profile(preferences_rk=[[0, 1], [1, 0]])
            >>> rule = RuleKimRoush()(profile)
            >>> rule.w_
            0
        """
        self.mylog("Count ballots", 1)
        scores = []
        candidates_worst_to_best = []
        one_v_might_be_pivotal = False
        # score_temp[c] = - number of vetos
        # score_temp[c] will be inf when c is eliminated
        score_temp = np.array(- np.bincount(
            self.profile_.preferences_rk[:, self.profile_.n_c - 1], minlength=self.profile_.n_c
        ).astype(float))
        score_average = - self.profile_.n_v / self.profile_.n_c
        candidates = np.array(range(self.profile_.n_c))
        is_alive = np.ones(self.profile_.n_c, dtype=np.bool)
        while True:  # This is a round
            scores.append(np.copy(score_temp))
            least_score = np.min(score_temp)
            self.mylogv("scores =", scores, 2)
            # self.mylogv("least_score =", least_score, 2)
            self.mylogv("score_average =", score_average, 2)
            if least_score == score_average:
                # Then all candidates have the average. This also includes the case where only one candidate is
                # remaining. We eliminate them by decreasing order of index (highest indexes first).
                losing_candidates = np.where(np.isfinite(score_temp))[0]
                candidates_worst_to_best.extend(losing_candidates[::-1])
                if len(losing_candidates) > 1:
                    one_v_might_be_pivotal = True
                    self.mylog("One voter might be pivotal", 2)
                self.mylogv("losing_candidates[::-1] =", losing_candidates[::-1], 2)
                break
            # Remove all candidates with score < average. Lowest score first, highest indexes first.
            while least_score < score_average:
                if least_score + 1 >= score_average:
                    one_v_might_be_pivotal = True
                    self.mylog("One voter might be pivotal", 2)
                losing_candidates = np.where(score_temp == least_score)[0]
                candidates_worst_to_best.extend(losing_candidates[::-1])
                self.mylogv("losing_candidates[::-1] =", losing_candidates[::-1], 2)
                is_alive[losing_candidates] = False
                score_temp[losing_candidates] = np.inf
                least_score = np.min(score_temp)
            if least_score - 1 < score_average:
                one_v_might_be_pivotal = True
                self.mylog("One voter might be pivotal", 2)
            candidates_alive = candidates[is_alive]
            score_temp = np.array(- np.bincount(
                candidates_alive[np.argmin(self.profile_.preferences_borda_rk[:, is_alive], 1)],
                minlength=self.profile_.n_c
            ).astype(float))
            score_temp[np.logical_not(is_alive)] = np.inf
            score_average = - self.profile_.n_v / np.sum(is_alive)
        candidates_by_scores_best_to_worst = np.array(candidates_worst_to_best[::-1])
        w = candidates_by_scores_best_to_worst[0]
        scores = np.array(scores)
        return {'scores': scores, 'w': w, 'one_v_might_be_pivotal': one_v_might_be_pivotal,
                'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[r, c]`` is minus the Veto score of candidate ``c`` at elimination round ``r``.

        By convention, if candidate ``c`` does not participate to round ``r``, then ``scores[r, c] = numpy.inf``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. Candidates are sorted according to their order of elimination. When several
        candidates are eliminated during the same round, they are sorted by Veto score at that round (more vetos are
        eliminated first) and, in case of a tie, by their index ( highest indexes are eliminated first).
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def v_might_im_for_c_(self):
        return np.full((self.profile_.n_v, self.profile_.n_c), self._count_ballots_['one_v_might_be_pivotal'])

    # %% CM

    def _cm_preliminary_checks_c_subclass_(self, c, optimize_bounds):
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        self._update_sufficient(self._sufficient_coalition_size_cm, c, (n_s + 1) * (self.profile_.n_c - 1),
                                'CM: Obvious bound => sufficient_coalition_size_cm[c] = ')



# -*- coding: utf-8 -*-
"""
Created on oct. 21, 2014, 09:54
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile
from svvamp.preferences.plurality_elimination_engine_profile_um import PluralityEliminationEngineProfileUM


class ProfileUM(Profile):

    def __init__(self, profile_s, n_m, ballot_ut=None, ballot_rk=None, ballot_borda_rk=None):
        self.profile_s = profile_s
        self.n_m = n_m
        if ballot_ut is None:
            self._ballot_ut_input = None
        else:
            self._ballot_ut_input = np.array(ballot_ut)
        if ballot_rk is None:
            self._ballot_rk_input = None
        else:
            self._ballot_rk_input = np.array(ballot_rk)
        if ballot_borda_rk is None:
            self._ballot_borda_rk_input = None
        else:
            self._ballot_borda_rk_input = np.array(ballot_borda_rk)
        super().__init__(preferences_ut=None, preferences_borda_rk=None, sort_voters=False)

    # %% Basic stuff

    @cached_property
    def n_c(self):
        return self.profile_s.n_c

    @cached_property
    def n_v(self):
        return self.profile_s.n_v + self.n_m

    @cached_property
    def labels_candidates(self):
        return self.profile_s.labels_candidates

    # %% Ballot of the manipulators

    @cached_property
    def ballot_rk(self):
        if self._ballot_rk_input is not None:
            return self._ballot_rk_input
        if self._ballot_borda_rk_input is not None:
            return np.array(sorted(range(self.n_c), key=self._ballot_borda_rk_input.__getitem__, reverse=True))
        if self._ballot_ut_input is not None:
            raise ValueError('You should not rely on the implicit conversion from utility to ranking'
                             'to compute UM.')
        raise ValueError('Please provide at least ballot_ut, ballot_rk or ballot_borda_rk.')

    @cached_property
    def ballot_borda_rk(self):
        if self._ballot_borda_rk_input is not None:
            return self._ballot_borda_rk_input
        else:
            return - np.array(np.argsort(self._ballot_rk_input), dtype=int) + self.n_c - 1

    @cached_property
    def ballot_ut(self):
        if self._ballot_ut_input is not None:
            return self._ballot_ut_input
        else:
            return self.ballot_borda_rk

    @cached_property
    def ballot_borda_ut(self):
        return np.array([
            np.sum(0.5 * (self.ballot_ut[c] >= self.ballot_ut) + 0.5 * (self.ballot_ut[c] > self.ballot_ut)) - 0.5
            for c in range(self.n_c)
        ])

    # %% Preferences

    @cached_property
    def preferences_ut(self):
        return np.concatenate((
            self.profile_s.preferences_ut,
            np.outer(np.ones(self.n_m, dtype=int), self.ballot_ut)
        ))

    @cached_property
    def preferences_borda_rk(self):
        return np.concatenate((
            self.profile_s.preferences_borda_rk,
            np.outer(np.ones(self.n_m, dtype=int), self.ballot_borda_rk)
        ))

    @cached_property
    def preferences_rk(self):
        return np.concatenate((
            self.profile_s.preferences_rk,
            np.outer(np.ones(self.n_m, dtype=int), self.ballot_rk)
        ))

    @cached_property
    def preferences_borda_ut(self):
        return np.concatenate((
            self.profile_s.preferences_borda_ut,
            np.outer(np.ones(self.n_m, dtype=int), self.ballot_borda_ut)
        ))

    # %% Matrix of duels

    @cached_property
    def matrix_duels_ut(self):
        return (
            self.profile_s.matrix_duels_ut
            + self.n_m * (self.ballot_ut[:, np.newaxis] > self.ballot_ut[np.newaxis, :])
        )

    @cached_property
    def matrix_duels_rk(self):
        return (
            self.profile_s.matrix_duels_rk
            + self.n_m * (self.ballot_borda_rk[:, np.newaxis] > self.ballot_borda_rk[np.newaxis, :])
        )

    @cached_property
    def plurality_scores_rk(self):
        plurality_scores_rk = np.zeros(self.n_c)
        plurality_scores_rk[self.ballot_rk[0]] = self.n_m
        plurality_scores_rk += self.profile_s.plurality_scores_rk
        return plurality_scores_rk

    @cached_property
    def plurality_scores_ut(self):
        plurality_scores_ut = np.zeros(self.n_c)
        preferred_candidates = np.where(self.ballot_ut == np.max(self.ballot_ut))[0]
        if preferred_candidates.size == 1:
            plurality_scores_ut[preferred_candidates[0]] = self.n_m
        plurality_scores_ut += self.profile_s.plurality_scores_ut
        return plurality_scores_ut

    def plurality_elimination_engine(self):
        return PluralityEliminationEngineProfileUM(self)



import numpy as np
from svvamp.preferences.plurality_elimination_engine import PluralityEliminationEngine


class PluralityEliminationEngineProfileUM(PluralityEliminationEngine):

    def __init__(self, profile):
        super().__init__(profile)

        self.preferences_borda_rk_s = profile.profile_s.preferences_borda_rk.copy()
        self.preference_borda_rk_m = profile.ballot_borda_rk.copy()

        self.ballots_s = np.argmax(self.preferences_borda_rk_s, axis=1)
        self.ballot_m = profile.ballot_rk[0]

        self._scores = profile.profile_s.plurality_scores_rk.copy().astype(float)
        self._scores[self.ballot_m] += profile.n_m

    def update_scores(self):
        self.preferences_borda_rk_s[:, self.loser] = -1
        new_ballots_s = np.argmax(self.preferences_borda_rk_s[self.ballots_s == self.loser, :], axis=1)
        self.ballots_s[self.ballots_s == self.loser] = new_ballots_s
        self._scores += np.bincount(new_ballots_s, minlength=self.profile.n_c)

        self.preference_borda_rk_m[self.loser] = -1
        if self.ballot_m == self.loser:
            self.ballot_m = np.argmax(self.preference_borda_rk_m)
            self.scores[self.ballot_m] += self.profile.n_m

        self.scores[self.loser] = np.nan

    @property
    def scores(self):
        return self._scores



from svvamp import ProfileFromFile
import os


def test_from_csv():
    file = os.path.join(os.path.dirname(__file__), 'example_ballots.csv')
    profile = ProfileFromFile(file, sort_voters=False)
    assert profile.preferences_ut.shape == (86, 11)
    assert len(profile.labels_candidates) == 11


def test_from_csv_transposed():
    file = os.path.join(os.path.dirname(__file__), 'example_ballots.t.csv')
    profile = ProfileFromFile(file, sort_voters=False)
    assert profile.preferences_ut.shape == (86, 11)
    assert len(profile.labels_candidates) == 11


def test_from_preflib():
    file = os.path.join(os.path.dirname(__file__), 'ED-00001-00000001.soi')
    profile = ProfileFromFile(file)
    assert profile.preferences_ut.shape == (43942, 12)
    assert profile.labels_candidates == [
        'Cathal Boland F.G.', 'Clare Daly S.P.', 'Mick Davis S.F.', 'Jim Glennon F.F.', 'Ciaran Goulding Non-P',
        'Michael Kennedy F.F.', 'Nora Owen F.G.', 'Eamonn Quinn Non-P', 'Sean Ryan Lab', 'Trevor Sargent G.P.',
        'David Henry Walshe C.C. Csp', 'G.V. Wright F.F.']


def test_from_preflib_with_ties():
    file = os.path.join(os.path.dirname(__file__), 'ED-00017-00000001.toi')
    profile = ProfileFromFile(file)
    assert profile.preferences_ut.shape == (4189, 4)



# -*- coding: utf-8 -*-
"""
Created on nov. 28, 2018, 14:55
Copyright François Durand 2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.preferences.generator_profile import GeneratorProfile
from svvamp.preferences.profile import Profile


class GeneratorProfileSpheroid(GeneratorProfile):
    """Profile generator using the 'Spheroid' model.

    Parameters
    ----------
    n_v : int
        Number of voters.
    n_c : int
        Number of candidates.
    stretching : number
        Number between 0 and ``numpy.inf`` (both included).
    sort_voters : bool
        This argument is passed to :class:`Profile`.

    Notes
    -----
    The utility vector of each voter is drawn independently and uniformly on a sphere in :math:`\\mathbb{R}^n_c`. Then,
    it is sent on the spheroid that is the image of the sphere by a dilatation of factor ``stretching`` in direction
    [1, ..., 1]. Cf. Durand et al., 'Geometry on the Utility Sphere'.

    The ordinal part of this distribution is the Impartial Culture.

    The parameter ``stretching`` has an influence only on voting systems based on utilities, especially Approval
    voting.

        * ``stretching = 0``: pure Von Neumann-Morgenstern utility, normalized to :math:`\\sum_c u_v(c) = 0` (
          spherical model with ``n_c-2`` dimensions).
       * ``stretching = 1``: spherical model with ``n_c-1`` dimensions.
        * ``stretching = inf``: axial/cylindrical model with only two possible values, all-approval [1, ...,
          1] and all-reject [-1, ..., -1].

    N.B.: This model gives is equivalent to a Von Mises-Fisher model with concentration = 0.

    Examples
    --------
    Typical usage:

        >>> generator = GeneratorProfileSpheroid(n_v=10, n_c=3, stretching=1)
        >>> profile = generator()
        >>> profile.preferences_rk.shape
        (10, 3)

    With some stretching:

        >>> generator = GeneratorProfileSpheroid(n_v=10, n_c=3, stretching=2)
        >>> profile = generator()
        >>> profile.preferences_rk.shape
        (10, 3)

        >>> generator = GeneratorProfileSpheroid(n_v=10, n_c=3, stretching=.5)
        >>> profile = generator()
        >>> profile.preferences_rk.shape
        (10, 3)
    """

    def __init__(self, n_v, n_c, stretching=1, sort_voters=False):
        self.n_v = n_v
        self.n_c = n_c
        self.stretching = stretching
        self.log_creation = ['Spheroid', n_c, n_v, 'Stretching', stretching]
        super().__init__(sort_voters=sort_voters)

    def __call__(self):
        # Step 1: spherical distribution
        preferences_ut = np.random.randn(self.n_v, self.n_c)
        preferences_ut = preferences_ut / np.sqrt(np.sum(preferences_ut**2, 1))[:, np.newaxis]
        # Step 2: apply stretching
        # These are not really different cases: same formula up to a factor 'stretching', which has no impact since we
        # work in a projective space. This distinction is only used to deal with stretching close or equal to 0 or
        # infinity. The matrix of the transformation is (up to a multiplicative factor):
        # (Id - 1/n_c J) + stretching * 1/n_c J,
        # where 1/n_c J is the orthogonal projection on diagonal (1, ..., 1) and (Id - 1/n_c J) is the projection on its
        # orthogonal hyperplane.
        if self.stretching < 1:
            preferences_ut = (preferences_ut
                              + np.sum(preferences_ut, 1)[:, np.newaxis] * (self.stretching - 1) / self.n_c)
        elif self.stretching > 1:
            preferences_ut = (preferences_ut / self.stretching
                              + np.sum(preferences_ut, 1)[:, np.newaxis] * (1 - 1 / self.stretching) / self.n_c)
        return Profile(preferences_ut=preferences_ut, log_creation=self.log_creation, sort_voters=self.sort_voters)



from svvamp.utils.misc import initialize_random_seeds
from svvamp import GeneratorProfileVMFHypersphere


def test():
    """
    When there is one group only, you may specify `vmf_probability` but it is ignored:

        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileVMFHypersphere(n_v=5, n_c=3, vmf_concentration=10, vmf_probability=1,
        ...                                            sort_voters=True)
        >>> profile = generator()
        >>> profile.preferences_ut
        array([[ 0.95891992, -0.12067454,  0.25672991],
               [ 0.86494714,  0.06265355,  0.49793673],
               [ 0.94128919, -0.27215259,  0.19976892],
               [ 0.80831535, -0.13959226,  0.57196179],
               [ 0.60411981, -0.07180745,  0.79365165]])
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on nov. 28, 2018, 13:42
Copyright François Durand 2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""


class GeneratorProfile:
    """A generator of profiles.

    Parameters
    ----------
    sort_voters : bool
        This argument is passed to :class:`Profile`.

    Examples
    --------
    Cf. :class:`GeneratorProfileEuclideanBox`.
    """

    def __init__(self, sort_voters=False):
        """Store the parameters of the generator.
        """
        self.sort_voters = sort_voters

    def __call__(self):
        """Return a random profile.

        Returns
        -------
        Profile
            A random profile.
        """
        raise NotImplementedError



import numpy as np
from svvamp.utils.misc import is_resistant_condorcet


def test():
    assert not is_resistant_condorcet(w=np.nan, preferences_ut=np.array([[5, 1, 2], [4, 10, 1]]))
    assert is_resistant_condorcet(w=0, preferences_ut=np.array([[5, 1, 2], [10, 4, 1]]))



# -*- coding: utf-8 -*-
"""
Created on 13 dec. 2018, 11:40
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import itertools
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.utils.pseudo_bool import equal_true, equal_false
from svvamp.preferences.profile import Profile
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut
from svvamp.rules.rule_irv import RuleIRV


class RuleCondorcetVtbIRV(Rule):
    """Condorcet Instant Runoff Voting

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleCondorcetVtbIRV()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [2. 0. 1.]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [2. 1. 0.]
        w = 0
        score_w = 2.0
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = fast, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 2. 4.]

    Notes
    -----
    Each voter must provide a strict total order. If there is a Condorcet winner (in the sense of
    :attr:`matrix_victories_rk`), then she is elected. Otherwise, :class:`RuleIRV` is used.

    If sincere preferences are strict total orders, then this voting system is equivalent to
    :class:`RuleCondorcetAbsIRV` for sincere voting, but manipulators have less possibilities (they are forced to
    provide strict total orders).

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Rely on :class:`RuleIRV`'s fast algorithm. Polynomial heuristic. Can prove
          CM but unable to decide non-CM (except in rare obvious cases).
        * :attr:`cm_option` = ``'slow'``: Rely on :class:`RuleExhaustiveBallot`'s exact algorithm. Non-polynomial
          heuristic (:math:`2^{n_c}`). Quite efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'very_slow'``: Rely on :class:`RuleIRV`'s exact algorithm. Non-polynomial
          heuristic (:math:`n_c!`). Very efficient to prove CM or non-CM.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

        Each algorithm above exploits the faster ones. For example, if :attr:`cm_option` = ``'very_slow'``,
        SVVAMP tries the fast algorithm first, then the slow one, then the 'very slow' one. As soon as it reaches
        a decision, computation stops.

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`. If
      :attr:`iia_subset_maximum_size` = 2, it runs in polynomial time and is exact up to ties (which can occur only if
      :attr:`n_v` is even).
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Condorcet criterion, ordinality and reduction of coalitional manipulability', François Durand,
    Fabien Mathieu and Ludovic Noirie, 2014.
    """

    full_name = 'Condorcet IRV'
    abbreviation = 'CIRV'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'cm_option': {'allowed': {'fast', 'slow', 'very_slow', 'exact'}, 'default': 'fast'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': {'exact'}, 'default': 'exact'}
    })

    def __init__(self, **kwargs):
        # self.irv_ = None
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_um=False, precheck_icm=False,
            log_identity="CONDORCET_VTB_IRV", **kwargs
        )

    def __call__(self, profile):
        self.delete_cache(suffix='_')
        self.profile_ = profile
        # Grab the IRV ballot of the profile (or create it)
        irv_options = {}
        if self.cm_option == 'fast':
            irv_options['cm_option'] = 'fast'
        elif self.cm_option == 'slow':
            irv_options['cm_option'] = 'slow'
        else:  # self.cm_option in {'very_slow', 'exact'}:
            irv_options['cm_option'] = 'exact'
        self.irv_ = RuleIRV(**irv_options)(self.profile_)
        return self

    # %% Counting the ballots

    @cached_property
    def w_(self):
        if not np.isnan(self.profile_.condorcet_winner_rk):
            return self.profile_.condorcet_winner_rk
        else:
            return self.irv_.w_

    @cached_property
    def scores_(self):
        """1d or 2d array.

            * If there is a Condorcet winner, then ``scores[c]`` is the number of victories for ``c`` in
              :attr:`matrix_duels_rk`.
            * Otherwise, ``scores[r, c]`` is defined like in :class:`RuleIRV`: it is the number of voters who vote
              for candidate ``c`` at round ``r``. For eliminated candidates, ``scores[r, c] = numpy.nan``. In contrast,
              ``scores[r, c] = 0`` means that ``c`` is present at round ``r`` but no voter votes for ``c``.
        """
        if not np.isnan(self.profile_.condorcet_winner_rk):
            return np.sum(self.profile_.matrix_victories_rk, 1)
        else:
            return self.irv_.scores_

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers.

            * If there is a Condorcet winner, candidates are sorted according to  their (scalar) score.
            * Otherwise, ``candidates_by_scores_best_to_worst`` is the list of all candidates in the reverse order of
              their IRV elimination.
        """
        if not np.isnan(self.profile_.condorcet_winner_rk):
            return np.argsort(- self.scores_, kind='mergesort')
        else:
            return self.irv_.candidates_by_scores_best_to_worst_

    @cached_property
    def _v_might_be_pivotal_(self):
        self.mylog("Compute _v_might_be_pivotal_", 1)
        if not np.isnan(self.profile_.condorcet_winner_rk):
            w = self.w_
            v_might_be_pivotal = np.zeros(self.profile_.n_v)
            for c in np.where(self.profile_.matrix_duels_rk[w, :] <= self.profile_.n_v / 2 + 1)[0]:
                if c == w:
                    continue
                # Search voters who can prevent the victory for ``w`` against ``c``.
                v_might_be_pivotal[
                    self.profile_.preferences_borda_rk[:, w] > self.profile_.preferences_borda_rk[:, c]
                ] = True
        else:
            eb = self.irv_.eb_
            # First way of being (maybe) pivotal: change the result of IRV.
            v_might_be_pivotal = np.copy(eb.v_might_be_pivotal_)
            # Another way of being (maybe) pivotal: create a Condorcet winner.
            for c in range(self.profile_.n_c):
                if np.any(self.profile_.matrix_duels_rk[:, c] >= self.profile_.n_v / 2 + 1):
                    # ``c`` cannot become a Condorcet winner.
                    continue
                # ``close_candidates`` are the candidates against which ``c`` does not have a victory.
                close_candidates = np.less_equal(self.profile_.matrix_duels_rk[c, :], self.profile_.n_v / 2)
                # Voter ``v`` can make ``c`` become a Condorcet winner iff, among ``close_candidates``, she vtb-likes
                # ``c`` the least (this way, she can improve ``c``'s scores in all these duels, compared to
                # her sincere voting).
                v_might_be_pivotal[np.all(np.less(
                        self.profile_.preferences_borda_rk[:, c][:, np.newaxis],
                        self.profile_.preferences_borda_rk[:, close_candidates]
                ), 1)] = True
        return v_might_be_pivotal

    @cached_property
    def v_might_im_for_c_(self):
        return np.tile(self._v_might_be_pivotal_[:, np.newaxis], (1, self.profile_.n_c))

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_majority_favorite_c_rk_ctb(self):
        return True

    @cached_property
    def meets_condorcet_c_rk(self):
        return True

    # %% Individual manipulation (IM)

    # TODO: should be implemented.

    # %% Trivial Manipulation (TM)

    # Use the general methods from class Rule.

    # %% Unison manipulation (UM)

    # TODO: should be implemented .

    def _um_preliminary_checks_general_subclass_(self):
        if equal_false(self.irv_.is_cm_):
            self._is_um = False
            self._candidates_um[:] = False
            self._um_was_computed_with_candidates = True

    def _um_preliminary_checks_c_(self, c):
        if self.um_option not in {'fast', 'lazy'} or self.cm_option not in {'fast', 'lazy'}:
            if (
                self.w_ == self.profile_.condorcet_winner_rk_ctb
                and not self.profile_.c_might_be_there_when_cw_is_eliminated_irv_style[c]
            ):
                # Impossible to manipulate with n_m manipulators
                self._candidates_um[c] = False

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods from class Rule.

    # %% Coalition Manipulation (CM)

    @property
    def losing_candidates_(self):
        """If ``irv_.w_ does not win, then we put her first. Other losers are sorted as usual. (scores in
        ``matrix_duels_ut``).

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.candidates_cm_
            array([ 0.,  1., nan])
        """
        self.mylog("Compute ordered list of losing candidates", 1)
        if self.w_ == self.irv_.w_:
            # As usual
            losing_candidates = np.concatenate((
                np.array(range(0, self.w_), dtype=int), np.array(range(self.w_ + 1, self.profile_.n_c), dtype=int)))
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
        else:
            # Put irv_.w_ first.
            losing_candidates = np.array(
                [c for c in range(self.profile_.n_c) if c != self.w_ and c != self.irv_.w_]).astype(np.int)
            losing_candidates = losing_candidates[np.argsort(
                - self.profile_.matrix_duels_ut[losing_candidates, self.w_], kind='mergesort')]
            losing_candidates = np.concatenate(([self.irv_.w_], losing_candidates))
        return losing_candidates

    def _cm_aux_(self, c, n_m, suggested_path, preferences_borda_s, matrix_duels_temp):
        """Algorithm used for CM (using suggested path).

        Parameters
        ----------
        c : int
            Candidate for which we want to manipulate.
        n_m : int
            The exact number of manipulators.
        suggested_path : List
            A suggested path of elimination (for the IRV part). It must work with ``n_m`` manipulators.
        preferences_borda_s : list of list (or ndarray)
            Preferences of the sincere voters (in Borda format).

        Returns
        -------
        bool
            ``manipulation_found``.

        Examples
        --------
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2, 3],
            ...     [0, 3, 1, 2],
            ...     [1, 3, 0, 2],
            ...     [2, 3, 1, 0],
            ...     [3, 2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV(cm_option='very_slow')(profile)
            >>> rule.candidates_cm_
            array([1., 1., 1., 0.])
        """
        candidates = np.array(range(self.profile_.n_c))

        # Step 1: elimination path
        # And consequences on the majority matrix
        scores_m_begin_r = np.zeros(self.profile_.n_c)
        is_candidate_alive_begin_r = np.ones(self.profile_.n_c, dtype=np.bool)
        current_top_v = np.array(- np.ones(n_m))  # -1 means that v is available
        candidates_to_put_in_ballot = np.ones((n_m, self.profile_.n_c), dtype=np.bool)
        for r in range(self.profile_.n_c - 1):
            self.mylogv("cm_aux: r =", r, 3)
            scores_tot_begin_r = np.full(self.profile_.n_c, np.nan)
            scores_tot_begin_r[is_candidate_alive_begin_r] = np.sum(np.equal(
                preferences_borda_s[:, is_candidate_alive_begin_r],
                np.max(preferences_borda_s[:, is_candidate_alive_begin_r], 1)[:, np.newaxis]
            ), 0)
            self.mylogv("cm_aux: scores_s_begin_r =", scores_tot_begin_r, 3)
            self.mylogv("cm_aux: scores_m_begin_r =", scores_m_begin_r, 3)
            scores_tot_begin_r += scores_m_begin_r
            self.mylogv("cm_aux: scores_tot_begin_r =", scores_tot_begin_r, 3)
            d = suggested_path[r]
            self.mylogv("cm_aux: d =", d, 3)
            scores_m_new_r = np.zeros(self.profile_.n_c, dtype=np.int)
            scores_m_new_r[is_candidate_alive_begin_r] = np.maximum(
                0,
                scores_tot_begin_r[d] - scores_tot_begin_r[is_candidate_alive_begin_r]
                + (candidates[is_candidate_alive_begin_r] > d))
            self.mylogv("cm_aux: scores_m_new_r =", scores_m_new_r, 3)
            # Update variables for next round
            scores_m_begin_r = scores_m_begin_r + scores_m_new_r
            if np.sum(scores_m_begin_r) > n_m:
                raise NotImplementedError("Error: this should not happen.")
            scores_m_begin_r[d] = 0
            is_candidate_alive_begin_r[d] = False
            # We need to attribute manipulator's votes to specific manipulators. This is done arbitrarily,
            # and has consequences on the majority matrix. It is the main cause that makes the algorithm not exact in
            #  theory (the other one being step 3).
            free_manipulators = np.where(current_top_v == -1)[0]
            i_manipulator = 0
            for e in range(self.profile_.n_c):
                n_manip_new_e = scores_m_new_r[e]
                for k in range(n_manip_new_e):
                    manipulator = free_manipulators[i_manipulator]
                    current_top_v[manipulator] = e
                    candidates_to_put_in_ballot[manipulator, e] = False
                    matrix_duels_temp[e, candidates_to_put_in_ballot[manipulator, :]] += 1
                    i_manipulator += 1
            current_top_v[current_top_v == d] = -1
            self.mylogv("cm_aux: current_top_v =", current_top_v, 3)
            self.mylogm("cm_aux: matrix_duels_temp =", matrix_duels_temp, 3)

        # Step 2
        # Ensure that no candidate ``!= c`` is Condorcet winner. If ``c`` is not yet in all ballots, put it.
        for manipulator in np.where(candidates_to_put_in_ballot[:, c])[0]:
            candidates_to_put_in_ballot[manipulator, c] = False
            matrix_duels_temp[c, candidates_to_put_in_ballot[manipulator, :]] += 1
        self.mylogv("cm_aux: Adding to all ballots c =", c, 3)
        self.mylogm("cm_aux: matrix_duels_temp =", matrix_duels_temp, 3)
        # If some candidates already have some non-victories in the matrix of duels, they can safely be put in the
        # ballots. Maybe this will generate non-victories for other candidates, etc.
        candidates_ok = np.zeros(self.profile_.n_c, dtype=np.bool)
        candidates_ok[c] = True
        i_found_a_new_ok = True
        while i_found_a_new_ok:
            not_yet_ok = np.where(np.logical_not(candidates_ok))[0]
            if not_yet_ok.shape[0] == 0:
                self.mylog("cm_aux: Decondorcification succeeded", 3)
                return True
            i_found_a_new_ok = False
            for d in not_yet_ok:
                if np.any(matrix_duels_temp[:, d] >= self.profile_.n_v / 2):
                    candidates_ok[d] = True
                    i_found_a_new_ok = True
                    for manipulator in np.where(candidates_to_put_in_ballot[:, d])[0]:
                        candidates_to_put_in_ballot[manipulator, d] = False
                        matrix_duels_temp[d, candidates_to_put_in_ballot[manipulator, :]] += 1
                    self.mylogv("cm_aux: Found a non-Condorcet d =", d, 3)
                    self.mylogm("cm_aux: matrix_duels_temp =", matrix_duels_temp, 3)

        # Step 3
        # Some candidates have left, who do not have non-victories yet. We will put them in the ballots,
        # while favoring a Condorcet cycle 0 > 1 > ... > C-1 > 0. N.B.: In practice, this step seems never necessary.
        self.mylog("cm_aux: Step 3 needed", 1)
        for manipulator in range(n_m):
            candidate_start = manipulator % self.profile_.n_c
            for d in itertools.chain(range(candidate_start, self.profile_.n_c), range(candidate_start)):
                if candidates_to_put_in_ballot[manipulator, d]:
                    candidates_to_put_in_ballot[manipulator, d] = False
                    matrix_duels_temp[d, candidates_to_put_in_ballot[manipulator, :]] += 1
        for d in np.where(np.logical_not(candidates_ok))[0]:
            if np.all(matrix_duels_temp[:, d] < self.profile_.n_v / 2):
                self.mylog("cm_aux: Decondorcification failed", 1)
                return False
        # TO DO: Investigate whether this case can actually happen.
        self._reached_uncovered_code()  # pragma: no cover
        self.mylog("cm_aux: Decondorcification succeeded", 1)  # pragma: no cover
        return True  # pragma: no cover

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV(cm_option='very_slow')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 3, 1],
            ...     [1, 2, 0, 3],
            ...     [1, 3, 0, 2],
            ...     [2, 0, 3, 1],
            ...     [3, 2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.candidates_cm_
            array([ 1., nan,  0., nan])

            >>> profile = Profile(preferences_ut=[
            ...     [-1. , -1. , -1. ],
            ...     [ 1. ,  1. , -0.5],
            ...     [-0.5,  1. , -1. ],
            ...     [-1. ,  1. ,  0. ],
            ...     [ 0.5, -0.5,  0.5],
            ... ], preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.candidates_cm_
            array([ 0., nan,  0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 1. , -0.5, -1. ],
            ...     [ 0.5, -1. ,  0.5],
            ...     [-1. ,  0.5,  0. ],
            ...     [-1. ,  1. , -1. ],
            ...     [ 1. , -1. ,  1. ],
            ... ], preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [1, 2, 0],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.candidates_cm_
            array([ 0., nan,  1.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.necessary_coalition_size_cm_
            array([2., 3., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 4, 3, 2, 1],
            ...     [1, 2, 3, 0, 4],
            ...     [2, 4, 3, 1, 0],
            ...     [3, 4, 1, 2, 0],
            ...     [4, 0, 1, 2, 3],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.necessary_coalition_size_cm_
            array([1., 2., 1., 1., 0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 1. ,  0. , -0.5],
            ...     [ 1. ,  0.5, -1. ],
            ...     [ 0. , -1. , -1. ],
            ...     [ 0.5, -0.5,  0.5],
            ...     [-0.5,  0.5,  1. ],
            ... ], preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV()(profile)
            >>> rule.necessary_coalition_size_cm_
            array([0., 3., 3.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV(cm_option='slow')(profile)
            >>> rule.necessary_coalition_size_cm_
            array([0., 5., 4.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleCondorcetVtbIRV(cm_option='exact')(profile)
            >>> rule.necessary_coalition_size_cm_
            array([3., 0., 3.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0.5, -0.5,  0. , -0.5, -1. ],
            ...     [ 0.5,  1. ,  1. , -1. ,  0. ],
            ...     [-1. ,  1. ,  0. ,  0. ,  0. ],
            ...     [ 0. ,  0. ,  1. ,  0.5, -1. ],
            ...     [ 0. , -1. ,  0. ,  0.5, -1. ],
            ...     [ 0.5,  0.5,  0. ,  0.5,  0.5],
            ... ], preferences_rk=[
            ...     [0, 2, 3, 1, 4],
            ...     [1, 2, 0, 4, 3],
            ...     [1, 3, 2, 4, 0],
            ...     [2, 3, 1, 0, 4],
            ...     [3, 2, 0, 4, 1],
            ...     [4, 3, 1, 0, 2],
            ... ])
            >>> rule = RuleCondorcetVtbIRV(cm_option='very_slow')(profile)
            >>> rule.necessary_coalition_size_cm_
            array([3., 3., 2., 0., 4.])
        """
        self.mylogv("CM: Compute CM for c =", c, 1)
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        candidates = np.array(range(self.profile_.n_c))
        preferences_borda_s = self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :]
        matrix_duels_vtb_temp = (preferences_ut_to_matrix_duels_ut(preferences_borda_s))
        self.mylogm("CM: matrix_duels_vtb_temp =", matrix_duels_vtb_temp, 3)
        # More preliminary checks. It's more convenient to put them in that method, because we need
        # ``preferences_borda_s`` and ``matrix_duels_vtb_temp``.
        d_neq_c = (np.array(range(self.profile_.n_c)) != c)
        n_manip_becomes_cond = np.maximum(n_s + 1 - 2 * np.min(matrix_duels_vtb_temp[c, d_neq_c]), 0)
        self.mylogv("CM: n_manip_becomes_cond =", n_manip_becomes_cond, 3)
        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_manip_becomes_cond,
                                'CM: Update sufficient_coalition_size_cm[c] = n_manip_becomes_cond =')
        if not optimize_bounds and n_m >= self._sufficient_coalition_size_cm[c]:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            return True
        # Prevent another cond. Look at the weakest duel for ``d``, she has ``matrix_duels_vtb_temp[d, e]``. We simply
        # need that:
        # ``matrix_duels_vtb_temp[d, e] <= (n_s + n_m) / 2``
        # ``2 * max_d(min_e(matrix_duels_vtb_temp[d, e])) - n_s <= n_m``
        n_manip_prevent_cond = 0
        for d in candidates[d_neq_c]:
            e_neq_d = (np.array(range(self.profile_.n_c)) != d)
            n_prevent_d = np.maximum(2 * np.min(matrix_duels_vtb_temp[d, e_neq_d]) - n_s, 0)
            n_manip_prevent_cond = max(n_manip_prevent_cond, n_prevent_d)
        self.mylogv("CM: n_manip_prevent_cond =", n_manip_prevent_cond, 3)
        self._update_necessary(self._necessary_coalition_size_cm, c, n_manip_prevent_cond,
                               'CM: Update necessary_coalition_size_cm[c] = n_manip_prevent_cond =')
        if not optimize_bounds and self._necessary_coalition_size_cm[c] > n_m:
            return True

        # Let us work
        if self.w_ == self.irv_.w_:
            self.mylog('CM: c != self.irv_.w == self.w', 3)
            if self.cm_option == "fast":
                self.irv_.cm_option = "fast"
            elif self.cm_option == "slow":
                self.irv_.cm_option = "slow"
            else:
                self.irv_.cm_option = "exact"
            # Use IRV without bounds
            self.mylog('CM: Use IRV without bounds')
            irv_is_cm_c = self.irv_.is_cm_c_(c)
            if equal_true(irv_is_cm_c):
                suggested_path_one = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_one, 3)
                manipulation_found = self._cm_aux_(c, n_m, suggested_path_one, preferences_borda_s,
                                                   matrix_duels_vtb_temp)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    if not optimize_bounds:
                        return True
            else:
                suggested_path_one = np.zeros(self.profile_.n_c)
            if equal_false(irv_is_cm_c):
                self.mylog('CM: irv_.is_cm_c_(c) = False', 3)
                self._update_necessary(self._necessary_coalition_size_cm, c,
                                       min(n_manip_becomes_cond, max(n_m + 1, n_manip_prevent_cond)),
                                       'CM: Update necessary_coalition_size[c] =')
                if not optimize_bounds:
                    return True
            if self._sufficient_coalition_size_cm[c] == self._necessary_coalition_size_cm[c]:
                return False
            # Use IRV with bounds. Either we have not decided manipulation for ``c``, or it is decided to False and
            # ``optimize_bounds = True`` (in that second case, we can only improve ``necessary_coalition_size[c]``).
            self.mylog('CM: Use IRV with bounds')
            self.irv_.is_cm_c_with_bounds_(c)
            self._update_necessary(
                self._necessary_coalition_size_cm, c,
                min(n_manip_becomes_cond, max(self.irv_.necessary_coalition_size_cm_[c], n_manip_prevent_cond)),
                'CM: Update necessary_coalition_size[c] =')
            if self.irv_.sufficient_coalition_size_cm_[c] <= n_m:
                suggested_path_two = self.irv_.example_path_cm_c_(c)
                self.mylogv("CM: suggested_path =", suggested_path_two, 3)
                if np.array_equal(suggested_path_one, suggested_path_two):
                    self.mylog('CM: Same suggested path as before, skip computation')
                else:
                    manipulation_found = self._cm_aux_(c, n_m, suggested_path_two, preferences_borda_s,
                                                       matrix_duels_vtb_temp)
                    self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                    if manipulation_found:
                        self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                                'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                        # We will not do better with exact (brute force) algorithm.
                        return False
        else:
            if c == self.irv_.w_:
                self.mylog('CM: c == self.irv_.w != self._w', 3)
                suggested_path = self.irv_.elimination_path_
                self.mylogv("CM: suggested_path =", suggested_path, 3)
                manipulation_found = self._cm_aux_(c, n_m, suggested_path, preferences_borda_s,
                                                   matrix_duels_vtb_temp)
                self.mylogv("CM: manipulation_found =", manipulation_found, 3)
                if manipulation_found:
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm[c] = n_m =')
                    # We will not do better with exact (brute force) algorithm.
                    return
            else:
                self.mylog('CM: c, self.irv_.w_ and self.w_ are all distinct', 3)
        if self.cm_option == 'exact':
            return self._cm_main_work_c_exact_(c, optimize_bounds)



# -*- coding: utf-8 -*-
"""
Created on 11 dec. 2018, 08:45
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
import networkx as nx
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.utils.misc import preferences_ut_to_matrix_duels_ut
from svvamp.utils.pseudo_bool import equal_true
from svvamp.preferences.profile import Profile


class RuleSchulze(Rule):
    """Schulze method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleSchulze()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[0 3 3]
         [2 0 2]
         [2 3 0]]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [[0 3 3]
         [2 0 3]
         [2 2 0]]
        w = 0
        score_w = [0 3 3]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (True)  ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (True)      ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = fast
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = fast
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = fast, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 2. 3.]

    Notes
    -----
    :attr:`scores_`\ ``[c, d]`` is equal to the width of the widest path from candidate ``c`` to candidate ``d`` in
    the capacited graph defined by :attr:`matrix_duels_rk`. We say that ``c`` is *better* than ``d`` if ``scores_[c,
    d]`` > ``scores_[d, c]``. Candidate ``c`` is a *potential winner* if no candidate ``d`` is *better* than ``c``.

    Among the potential winners, the candidate with lowest index is declared the winner.

    .. note::

        In the original Schulze method, ties are broken at random. However, this feature is not supported by SVVAMP
        because it leads to difficulties for the *definition* of manipulation itself (and all the more for its
        implementation).

    * :meth:`is_cm_`:

        * :attr:`cm_option` = ``'fast'``: Gaspers et al. (2013). This algorithm is polynomial and has a window of
          error of 1 manipulator (due to the tie-breaking rule).
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`:

        * :attr:`im_option` = ``'fast'``: Gaspers et al. (2013). This algorithm is polynomial and may not be able to
          decide IM (due to the tie-breaking rule).
        * :attr:`im_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

    * :meth:`~svvamp.Election.not_iia`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`:

        * :attr:`um_option` = ``'fast'``: Gaspers et al. (2013). This algorithm is polynomial and has a window of
          error of 1 manipulator (due to the tie-breaking rule).
        * :attr:`um_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

    .. note:

        For this voting system, UM and CM are almost equivalent up to tie-breaking. For this reason,
        :attr:`um_option` and :attr:`cm_option` are linked to each other: modifying one modifies the other accordingly.

    References
    ----------
    'A new monotonic, clone-independent, reversal symmetric, and Condorcet-consistent single-winner election
    method ', Markus Schulze, 2011.

    'Schulze and Ranked-Pairs Voting are Fixed-Parameter Tractable to Bribe, Manipulate, and Control',
    Lane A. Hemaspaandra, Rahman Lavaee and Curtis Menton, 2012.

    'Manipulation and Control Complexity of Schulze Voting', Curtis Menton and Preetjot Singh, 2012.

    'A Complexity-of-Strategic-Behavior Comparison between Schulze’s Rule and Ranked Pairs', David Parkes and
    Lirong Xia, 2012.

    'Coalitional Manipulation for Schulze’s Rule', Serge Gaspers, Thomas Kalinowski, Nina Narodytska and Toby
    Walsh, 2013.
    """

    full_name = 'Schulze'
    abbreviation = 'Sch'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'im_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
        'icm_option': {'allowed': ['exact'], 'default': 'exact'},
        'cm_option': {'allowed': ['fast', 'exact'], 'default': 'fast'}
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_um=False, precheck_icm=False,
            log_identity="SCHULZE", **kwargs
        )

    # %% Counting the ballots

    @staticmethod
    def _count_ballot_aux(matrix_duels):
        n_c = matrix_duels.shape[0]
        # Calculate widest path for candidate c to d
        widest_path = np.copy(matrix_duels)
        for i in range(n_c):
            for j in range(n_c):
                if i == j:
                    continue
                for k in range(n_c):
                    if k == i or k == j:
                        continue
                    widest_path[j, k] = max(widest_path[j, k], min(widest_path[j, i], widest_path[i, k]))
        # ``c`` is better than ``d`` if ``widest_path[c, d] > widest_path[d, c]``, or if there is equality and
        # ``c < d``. Potential winners: maximal ``better_or_tie[c]``. Tie-break: index of the candidate. N.B.: if we
        # broke ties with ``better_c = sum(... > ...)``, it would not meet ``condorcet_c_vtb_ctb``.
        better_or_tie_c = np.sum(widest_path >= widest_path.T, 1)
        candidates_by_scores_best_to_worst = np.argsort(- better_or_tie_c, kind='mergesort')
        w = candidates_by_scores_best_to_worst[0]
        return {'w': w, 'scores': widest_path, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        return self._count_ballot_aux(self.profile_.matrix_duels_rk)

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. ``candidates_by_scores_best_to_worst_[k]`` is the ``k``\ :sup:`th` candidate by
        number of Schulze-victories, i.e. the number of candidates ``d`` such that ``c`` is *better* than ``d``.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[c, d]`` is equal to the width of the widest path from ``c`` to ``d``.

        .. note::

            Unlike for most other voting systems, ``scores`` matrix must be read in rows, in order to comply with our
            convention for the matrix of duels: ``c``'s score vector is ``scores[c, :]``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def score_w_(self):
        """1d array. ``score_w_`` is :attr:`w_`'s score vector: ``score_w_`` =
        :attr:`scores_`\ ``[``:attr:`w_`\ ``, :]``.
        """
        self.mylog("Compute winner's score", 1)
        return self.scores_[self.w_, :]

    @cached_property
    def scores_best_to_worst_(self):
        """2d array. ``scores_best_to_worst`` is the scores of the candidates, from the winner to the last candidate
        of the election.

        ``scores_best_to_worst[k, j]`` is the width of the widest path from the ``k``\ :sup:`th` best candidate of
        the election to the ``j``\ :sup:`th`.
        """
        self.mylog("Compute scores_best_to_worst", 1)
        return self.scores_[self.candidates_by_scores_best_to_worst_, :][:, self.candidates_by_scores_best_to_worst_]

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        return True

    # %% Setting the options

    @property
    def um_option(self):
        return self._um_option

    @um_option.setter
    def um_option(self, value):
        if self._um_option == value:
            return
        if value in self.options_parameters['um_option']['allowed']:
            self.mylogv("Setting um_option =", value, 1)
            self._cm_option = value
            self._um_option = value
            self.delete_cache(contains='_um_', suffix='_')
            self.delete_cache(contains='_cm_', suffix='_')
        else:
            raise ValueError("Unknown value for um_option: " + format(value))

    @property
    def cm_option(self):
        return self._cm_option

    @cm_option.setter
    def cm_option(self, value):
        if self._cm_option == value:
            return
        if value in self.options_parameters['cm_option']['allowed']:
            self.mylogv("Setting cm_option =", value, 1)
            self._cm_option = value
            self._um_option = value
            self.delete_cache(contains='_um_', suffix='_')
            self.delete_cache(contains='_cm_', suffix='_')
        else:
            raise ValueError("Unknown value for cm_option: " + format(value))

    # %% Manipulation routine

    def _vote_strategically(self, matrix_duels_s, strength, c, n_m):
        """ Manipulation algorithm: "Coalitional Manipulation for Schulze’s Rule" (Gaspers, Kalinowski, Narodytska
        and Walsh 2013).

        Parameters
        ----------
        matrix_duels_s : ndarray
            2d array. Matrix of duels for sincere voters.
        strength : ndarray
            2d array. ``S[c, d]`` is the strength of the widest path from ``c`` to ``d`` (with sincere
            voters only). Be careful, it is based on the antisymmetric matrix of duels. I.e. while
            ``_count_ballot_aux`` provides ``widest_path_s``, we need to take ``S = 2 * (widest_path_s - n_s / 2)``,
            then set ``S``'s diagonal coefficients to 0.
        c : int
            Candidate. Our challenger.
        n_m : int
            Number of manipulators.
        """
        candidates_not_c = np.concatenate((np.array(range(c), dtype=int),
                                           np.array(range(c + 1, self.profile_.n_c), dtype=int)))
        w = matrix_duels_s - matrix_duels_s.T
        w_sup = w + n_m
        w_inf = w - n_m
        matrix_u = strength + n_m
        matrix_u_prev = None
        self.mylogv("candidates_not_c =", candidates_not_c, 3)
        self.mylogm("matrix_duels_s =", matrix_duels_s, 3)
        self.mylogm("S =", strength, 3)
        self.mylogm("w =", w, 3)
        self.mylogm("w_sup =", w_sup, 3)
        self.mylogm("w_inf =", w_inf, 3)
        self.mylogm("matrix_u =", matrix_u, 3)
        self.mylogm("matrix_u_prev =", matrix_u_prev, 3)
        # Algorithm: Pre-processing Bounds
        while not np.all(matrix_u_prev == matrix_u):
            matrix_u_prev = np.copy(matrix_u)
            # Rule 1
            matrix_u[:, c] = np.minimum(matrix_u[:, c], matrix_u[c, :])
            self.mylogm('Rule 1: U =', matrix_u, 3)
            # Rule 2
            for x in candidates_not_c:
                matrix_g_x = np.ones((self.profile_.n_c, self.profile_.n_c))
                # Remove all vertices y (candidates) s.t. U(y,c) < U(x,c)
                matrix_v_removed = np.less(matrix_u[:, c], matrix_u[x, c])
                matrix_v_removed[c] = False
                matrix_g_x[matrix_v_removed, :] = 0
                matrix_g_x[:, matrix_v_removed] = 0
                # Removed all edges (y,z) s.t. w_sup(y,z) < U(x,c)
                matrix_g_x[w_sup < matrix_u[x, c]] = 0
                # Does matrix_g_x contain a path from c to x ? If not, do stuff.
                if not nx.has_path(nx.DiGraph(matrix_g_x), c, x):
                    matrix_u[x, c] -= 2
                    self.mylogm('Rule 2: U =', matrix_u, 3)
            # Rule 3
            for x in candidates_not_c:
                for y in candidates_not_c:
                    if y == x:
                        continue
                    if matrix_u[x, c] < w_inf[x, y]:
                        matrix_u[y, c] = min(matrix_u[y, c], matrix_u[x, c])
                        self.mylogm('Rule 3: U =', matrix_u, 3)
            # Test possibility
            self.mylogm("U =", matrix_u, 3)
            self.mylogv("U[candidates_not_c, c] =", matrix_u[candidates_not_c, c], 3)
            self.mylogv("S[candidates_not_c, c] - n_m =", strength[candidates_not_c, c] - n_m, 3)
            if np.any(matrix_u[candidates_not_c, c] < strength[candidates_not_c, c] - n_m):
                self.mylog("Cowinner manipulation failed")
                return False, False
        # Algorithm : construction of ordering Lambda (when manipulation is possible)
        array_f = np.zeros(self.profile_.n_c, dtype=np.bool)
        array_f[c] = 1
        array_x = np.ones(self.profile_.n_c, dtype=np.bool)
        array_x[c] = 0
        order_lambda = [c]
        for i in range(self.profile_.n_c - 1):
            the_d = np.max(matrix_u[array_x, c])
            possible_y = np.logical_and(np.logical_and(array_x, matrix_u[:, c] == the_d),
                                        np.any(w_sup[array_f, :] >= the_d, 0))
            y = np.where(possible_y)[0][-1]
            array_f[y] = True
            array_x[y] = False
            order_lambda.append(y)
        order_lambda = np.array(order_lambda)
        self.mylogv("order_lambda =", order_lambda, 3)
        # Who wins with the tie-breaking rule?
        reciprocal_lambda = np.argsort(order_lambda)
        matrix_duels_m = n_m * np.triu(
            np.ones((self.profile_.n_c, self.profile_.n_c)), 1
        )[reciprocal_lambda, :][:, reciprocal_lambda]
        w_temp = self._count_ballot_aux(matrix_duels_s + matrix_duels_m)['w']
        self.mylogv("matrix_duels_m =", matrix_duels_m, 3)
        self.mylogv("w_temp =", w_temp, 3)
        if w_temp == c:
            self.mylog("Cowinner manipulation worked, with ctb also")
            return True, True
        else:
            self.mylog("Cowinner manipulation worked but not with ctb")
            return True, False

    # %% Individual manipulation (IM)

    def _im_main_work_v_fast(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.is_im_
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.is_im_c_(2)
            False

            >>> profile = Profile(preferences_rk=[
            ...     [1, 3, 2, 0, 4],
            ...     [2, 3, 0, 1, 4],
            ...     [3, 4, 2, 1, 0],
            ...     [4, 0, 1, 3, 2],
            ...     [4, 2, 0, 1, 3],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.is_im_c_(2)
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 3, 4, 2],
            ...     [2, 4, 0, 3, 1],
            ...     [2, 4, 1, 3, 0],
            ...     [3, 2, 0, 1, 4],
            ...     [4, 0, 3, 2, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.v_im_for_c_
            array([[ 0.,  0.,  0.,  0.,  0.],
                   [ 0.,  0.,  1.,  0.,  1.],
                   [ 0.,  0., nan, nan,  1.],
                   [ 0.,  0.,  1., nan,  0.],
                   [ 0.,  0.,  0.,  0., nan]])
        """
        for c in self.losing_candidates_:
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            self.mylogv("IM: c =", c, 3)
            nb_wanted_undecided -= 1
            # Maybe we will not decide, but we will have done the 'fast' job for ``c`` anyway.

            matrix_duels_s = np.copy(self.profile_.matrix_duels_rk)
            for x in range(self.profile_.n_c):
                for y in range(x + 1, self.profile_.n_c):
                    if self.profile_.preferences_borda_rk[v, x] > self.profile_.preferences_borda_rk[v, y]:
                        matrix_duels_s[x, y] -= 1
                    else:
                        matrix_duels_s[y, x] -= 1

            count_ballot_s = self._count_ballot_aux(matrix_duels_s)
            w_s, widest_path = count_ballot_s['w'], count_ballot_s['scores']
            if w_s == c:
                self.mylog("IM: Manipulation easy (c wins without manipulator's vote)", 3)
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true or nb_wanted_undecided == 0:
                    return
                continue  # pragma: no cover - Not really "executed" because of Python optimization

            the_s = 2 * (widest_path - (self.profile_.n_v - 1) / 2)
            np.fill_diagonal(the_s, 0)
            success_cowinner, success_tb = self._vote_strategically(matrix_duels_s, the_s, c, 1)
            if success_tb:
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true:
                    return
            elif success_cowinner:
                self.mylog("IM not sure", 3)
                if self._im_option != 'exact':
                    self._v_im_for_c[v, c] = np.nan
            else:
                self._v_im_for_c[v, c] = False
            if nb_wanted_undecided == 0:
                return

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze(im_option='exact')(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleSchulze(im_option='exact')(profile)
            >>> rule.is_im_
            False

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze(im_option='exact')(profile)
            >>> rule.is_im_
            True
        """
        self._im_main_work_v_fast(v, c_is_wanted, nb_wanted_undecided, stop_if_true)
        # Deal with 'exact' (brute force) option
        if self._im_option != 'exact':
            return
        if stop_if_true and np.any(np.equal(self._v_im_for_c[v, c_is_wanted], True)):
            return
        nb_wanted_undecided_updated = np.sum(np.isneginf(self._v_im_for_c[v, c_is_wanted]))
        if nb_wanted_undecided_updated == 0:
            self.mylog("IM: Job finished", 3)
        else:
            self.mylogv("IM: Still some work for v =", v, 3)
            self._im_main_work_v_exact_(v, c_is_wanted, nb_wanted_undecided_updated, stop_if_true)

    # %% Trivial Manipulation (TM)

    # Use the general methods.

    # %% Ignorant-Coalition Manipulation (ICM)

    # Use the general methods.

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 3, 1, 2],
            ...     [1, 0, 2, 3],
            ...     [2, 1, 3, 0],
            ...     [2, 3, 0, 1],
            ...     [3, 2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_cm_
            array([ 1.,  1.,  0., nan])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 3, 2, 1],
            ...     [1, 0, 3, 2],
            ...     [2, 1, 0, 3],
            ...     [2, 3, 0, 1],
            ...     [3, 1, 0, 2],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_cm_
            array([ 0.,  1., nan,  1.])

            >>> profile = Profile(preferences_rk=[
            ...     [1, 0, 2, 3, 4],
            ...     [2, 1, 4, 0, 3],
            ...     [2, 3, 4, 0, 1],
            ...     [3, 4, 0, 1, 2],
            ...     [4, 0, 3, 1, 2],
            ...     [4, 0, 3, 1, 2],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_cm_
            array([0., 0., 1., 1., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleSchulze(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0. ,  1. ,  1. , -1. ,  0. ],
            ...     [-0.5,  1. ,  0. ,  0.5,  0.5],
            ...     [ 1. , -0.5,  0.5,  1. , -0.5],
            ...     [ 0. ,  0.5,  0.5,  1. , -0.5],
            ...     [-1. , -1. , -0.5,  0.5,  1. ],
            ...     [-0.5, -0.5,  0. ,  0.5,  1. ],
            ... ], preferences_rk=[
            ...     [1, 2, 4, 0, 3],
            ...     [1, 3, 4, 2, 0],
            ...     [3, 0, 2, 4, 1],
            ...     [3, 2, 1, 0, 4],
            ...     [4, 3, 2, 0, 1],
            ...     [4, 3, 2, 1, 0],
            ... ])
            >>> rule = RuleSchulze(cm_option='exact')(profile)
            >>> rule.is_um_c_(0)
            False
            >>> rule.necessary_coalition_size_cm_
            array([2., 2., 2., 0., 2.])
        """
        _ = self._um_variables_are_declared_
        matrix_duels_s = preferences_ut_to_matrix_duels_ut(
            self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :])
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        n_s = self.profile_.n_v - n_m
        count_ballot_s = self._count_ballot_aux(matrix_duels_s)
        w, widest_path = count_ballot_s['w'], count_ballot_s['scores']
        the_s = 2 * (widest_path - n_s / 2)
        np.fill_diagonal(the_s, 0)

        # We try the "co-winner" version with ``n_m - 1`` manipulators. If True, it implies that unique winner
        # version is OK for ``n_m`` manipulators.
        success_cowinner, success_tb = self._vote_strategically(matrix_duels_s, the_s, c, n_m - 1)
        if success_tb:
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m - 1,
                                    'CM: Update sufficient_coalition_size_cm = n_m - 1 =')
            self._candidates_um[c] = True
            self._is_um = True
            return
        if success_cowinner:
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                    'CM: Update sufficient_coalition_size_cm = n_m =')
            self._update_necessary(self._necessary_coalition_size_cm, c, n_m - 1,
                                   'CM: Update necessary_coalition_size_cm = n_m - 1 =')
            # We do not know for UM, but we do not really care
            return
        else:
            self._update_necessary(self._necessary_coalition_size_cm, c, n_m,
                                   'CM: Update necessary_coalition_size_cm = n_m =')

        # We try with n_m manipulators.
        success_cowinner, success_tb = self._vote_strategically(matrix_duels_s, the_s, c, n_m)
        self._um_fast_tested[c] = True
        if success_tb:
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                    'CM: Update sufficient_coalition_size_cm = n_m =')
            self._candidates_um[c] = True
            self._is_um = True
            return
        if success_cowinner:
            self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m + 1,
                                    'CM: Update sufficient_coalition_size_cm = n_m + 1 =')
            self._update_necessary(self._necessary_coalition_size_cm, c, n_m,
                                   'CM: Update necessary_coalition_size_cm = n_m =')
            if self.cm_option == 'exact':
                self._um_main_work_c_exact_rankings_(c)
                if equal_true(self._candidates_um[c]):
                    self._update_sufficient(self._sufficient_coalition_size_cm, c, n_m,
                                            'CM: Update sufficient_coalition_size_cm = n_m =')
                else:
                    self._cm_main_work_c_exact_(c, optimize_bounds)
        else:
            self._update_necessary(self._necessary_coalition_size_cm, c, n_m + 1,
                                   'CM: Update necessary_coalition_size_cm = n_m + 1 =')
            self._candidates_um[c] = False
            return

    # %% Unison manipulation (UM)

    @cached_property
    def _um_variables_are_declared_(self):
        self._um_was_computed_with_candidates = False
        # Exceptionally, we initialize the variables for UM here, because UM is in great part managed in CM.
        self._is_um = - np.inf
        self._candidates_um = np.full(self.profile_.n_c, -np.inf)
        self._candidates_um[self.w_] = False
        # ``_um_fast_tested[c]`` will be True iff if we have already launched ``_vote_strategically`` with ``n_m``
        # manipulators (for ``c``).
        self._um_fast_tested = np.zeros(self.profile_.n_c, dtype=np.bool)
        return True

    @cached_property
    def _um_is_initialized_general_(self):
        self.mylog("UM: Initialize", 2)
        _ = self._um_variables_are_declared_
        self._um_preliminary_checks_general_()
        return True

    def _um_preliminary_checks_general_subclass_(self):
        if not np.any(np.isneginf(self._candidates_um)):
            return
        _ = self.is_cm_
        self._candidates_um[np.equal(self._candidates_cm, False)] = False

    def _um_main_work_c_(self, c):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_um_
            array([0., 1., 1.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 1. ,  1. ,  0. ,  0.5, -0.5],
            ...     [ 0. , -0.5,  1. , -0.5,  0. ],
            ...     [ 0. ,  0. ,  1. ,  1. ,  1. ],
            ...     [ 0.5, -0.5, -0.5, -0.5,  0.5],
            ...     [ 0.5, -0.5, -1. ,  0.5,  1. ],
            ...     [-1. ,  0.5, -1. ,  0.5,  1. ],
            ... ], preferences_rk=[
            ...     [1, 0, 3, 2, 4],
            ...     [2, 4, 0, 1, 3],
            ...     [3, 2, 4, 1, 0],
            ...     [4, 0, 2, 1, 3],
            ...     [4, 0, 3, 1, 2],
            ...     [4, 1, 3, 0, 2],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_um_
            array([0., 0., 1., 0., 0.])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleSchulze()(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])
            >>> rule.candidates_um_
            array([0., 0., 0.])

            >>> profile = Profile(preferences_ut=[
            ...     [ 0. ,  0.5,  0. , -1. ],
            ...     [ 0. , -0.5,  0.5, -0.5],
            ... ], preferences_rk=[
            ...     [1, 2, 0, 3],
            ...     [2, 0, 3, 1],
            ... ])
            >>> rule = RuleSchulze(um_option='exact', cm_option='exact')(profile)
            >>> rule.candidates_um_
            array([1., 0., 0., 0.])
        """
        n_m = self.profile_.matrix_duels_ut[c, self.w_]
        if self._sufficient_coalition_size_cm[c] + 1 <= n_m:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            self._candidates_um[c] = True
            return
        if self._necessary_coalition_size_cm[c] - 1 > n_m:  # pragma: no cover
            # TO DO: Investigate whether this case can actually happen.
            self._reached_uncovered_code()
            self._candidates_um[c] = False
            return
        if not self._um_fast_tested[c]:
            # We must try the fast algo first.
            matrix_duels_s = preferences_ut_to_matrix_duels_ut(
                self.profile_.preferences_borda_rk[np.logical_not(self.v_wants_to_help_c_[:, c]), :])
            n_s = self.profile_.n_v - n_m
            count_ballot_s = self._count_ballot_aux(matrix_duels_s)
            w, widest_path = count_ballot_s['w'], count_ballot_s['scores']
            the_s = 2 * (widest_path - n_s / 2)
            np.fill_diagonal(the_s, 0)
            success_cowinner, success_tb = self._vote_strategically(matrix_duels_s, the_s, c, n_m)
            self._um_fast_tested[c] = True
            if success_tb:
                self._candidates_um[c] = True
                return
            if not success_cowinner:
                self._candidates_um[c] = False
                return
        if self.um_option == 'exact':
            self._um_main_work_c_exact_rankings_(c)
        else:
            self._candidates_um[c] = np.nan



# -*- coding: utf-8 -*-
"""
Created on jul. 26, 2021, 08:43
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pathlib import Path
import tikzplotlib


class ExperimentsCompiler:

    def __init__(self, prefix_tikz_file, tikz_directory='tikz', results_directory='out', figsize=(16, 8)):
        # Record parameters
        self.prefix_tikz_file = prefix_tikz_file
        self.results_directory = Path(results_directory)
        self.tikz_directory = Path(tikz_directory)
        self.figsize = figsize
        # Create the mother of all dataframes
        dataframes = []
        for f in self.results_directory.iterdir():
            if '_m_en_' not in f.name:
                continue
            with f.open() as buffer:
                try:
                    new_df = pd.read_csv(buffer)
                except UnicodeDecodeError:
                    print(f)
                    raise UnicodeDecodeError
            new_df['file_name'] = f.name
            dataframes.append(new_df)
        self.df = pd.concat(dataframes, ignore_index=True)
        # Default order of the rules
        df_temp = self.df[self.df['Criterion'] == 'Winner'].pivot_table(
            values=['Rate (lower bound)'],
            index='Rule (abbr)'
        )

        def order_rules(rule_abbreviation):
            rules = [
                'Ben', 'SIRV', 'Tid', 'Woo',
                'CIRV', 'IRV', 'EB',
                'STAR', 'AV', 'RV'
            ]
            try:
                # noinspection PyUnresolvedReferences
                return rules.index(rule_abbreviation)
            except ValueError:
                return len(rules) + 1

        self.rules_order = sorted(list(df_temp.index), key=order_rules)
        # Number of rules
        self.n_rules = len(self.rules_order)

    def profiles_scatter_plot(self, tikz_file='profiles_scatter_plot.tex'):
        df_scatter = self.df[
            self.df['Criterion'] == 'exists_condorcet_admissible'  # No matter the criterion
            ]
        plt.subplots(figsize=self.figsize)
        plt.scatter(x=df_scatter['V'], y=df_scatter['C'])
        plt.xlabel('Number of voters V')
        plt.xscale('log')
        plt.ylabel('Number of candidates C')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_scatter

    def profile_features_bar_plot(self, tikz_file='profile_features_bar_plot.tex'):
        d_criterion_legend = {
            'exists_condorcet_winner_rk': 'Condorcet winner',
            'exists_condorcet_order_rk': 'Condorcet order',
            'exists_irv_immune_candidate': 'IRV-immune candidate',
            'exists_resistant_condorcet_winner': 'Resistant Condorcet winner',
            'exists_majority_favorite_rk': 'Majority favorite'
        }
        d_criterion_rate = {}
        for criterion in d_criterion_legend.keys():
            d_criterion_rate[criterion] = self.df[
                self.df['Criterion'] == criterion
                ].pivot_table(
                values=['Rate (lower bound)'],
                index='Rule'
            ).iloc[0, 0]
        x = list(d_criterion_legend.values())
        y = [d_criterion_rate[criterion] for criterion in d_criterion_legend.keys()]
        df_plot = pd.DataFrame(y, index=x, columns=['Rate'])
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        plt.bar(x, y)
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('Proportion of profiles')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_plot

    def rate_bar_plot(self, criterion, ylabel, tikz_file):
        # Create the pivot table
        df_plot = self.df[self.df['Criterion'] == criterion].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)'
        )
        df_plot = df_plot.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]
        df_plot.sort_values(by=['Rate (lower bound)', 'Rate (upper bound)', ], inplace=True)
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        plt.bar(
            df_plot.index,
            df_plot['Rate (lower bound)'],
            yerr=(np.zeros(len(df_plot.index)), df_plot['Rate (uncertainty)'])
        )
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel(ylabel)
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_plot

    def cm_rate_bar_plot(self, tikz_file='cm_rate_bar_plot.tex'):
        return self.rate_bar_plot(criterion='is_cm_', ylabel='CM rate', tikz_file=tikz_file)

    def tm_rate_bar_plot(self, tikz_file='tm_rate_bar_plot.tex'):
        return self.rate_bar_plot(criterion='is_tm_', ylabel='TM rate', tikz_file=tikz_file)

    def um_rate_bar_plot(self, tikz_file='um_rate_bar_plot.tex'):
        return self.rate_bar_plot(criterion='is_um_', ylabel='UM rate', tikz_file=tikz_file)

    def cm_tm_um_rate_bar_plot(self, tikz_file='cm_tm_um_rate_bar_plot.tex'):
        # Create the pivot table
        df_tm = self.df[self.df['Criterion'] == 'is_tm_'].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)'
        )
        df_tm = df_tm.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]

        df_um = self.df[self.df['Criterion'] == 'is_um_'].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)'
        )
        df_um = df_um.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]

        df_cm = self.df[self.df['Criterion'] == 'is_cm_'].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)'
        )
        df_cm = df_cm.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]

        df_plot = pd.concat([df_tm, df_um, df_cm], axis=1, keys=['TM', 'UM', 'CM'])
        df_plot.sort_values(by=[
            ('CM', 'Rate (lower bound)'), ('CM', 'Rate (upper bound)'),
            ('TM', 'Rate (lower bound)'), ('TM', 'Rate (upper bound)'),
            ('UM', 'Rate (lower bound)'), ('UM', 'Rate (upper bound)'),
        ], inplace=True)
        # Plot
        x = np.arange(len(df_plot.index))
        bar_width = 0.25
        fig, ax = plt.subplots(figsize=self.figsize)
        ax.bar(x - bar_width, df_plot[('TM', 'Rate (lower bound)')], bar_width, label='TM rate',
               yerr=(np.zeros(len(df_plot.index)), df_plot[('TM', 'Rate (uncertainty)')]))
        ax.bar(x, df_plot[('UM', 'Rate (lower bound)')], bar_width, label='UM rate',
               yerr=(np.zeros(len(df_plot.index)), df_plot[('UM', 'Rate (uncertainty)')]))
        ax.bar(x + bar_width, df_plot[('CM', 'Rate (lower bound)')], bar_width, label='CM rate',
               yerr=(np.zeros(len(df_plot.index)), df_plot[('CM', 'Rate (uncertainty)')]))
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        ax.set_xticks(x)
        ax.set_xticklabels(df_plot.index)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('Rate')
        ax.legend(loc='lower right')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_plot

    def condorcet_consistency_bar_plot(self, tikz_file='condorcet_consistency_bar_plot.tex'):
        # Cumulative rate of Condorcet winner existence
        cum_rate_condorcet_exists = self.df[
            (self.df['Rule'] == 'Profile') & (self.df['Criterion'] == 'exists_condorcet_winner_rk')
            ].pivot_table(
            values='Rate (lower bound)',
            index='Rule',
            aggfunc=np.sum
        ).iloc[0, 0]
        # Sincere Condorcet consistency
        df_condorcet_consistency = self.df[
            self.df['Criterion'] == 'w_is_condorcet_winner_rk_'
            ].pivot_table(
            values='Rate (lower bound)',
            index='Rule (abbr)',
            aggfunc=np.sum
        )
        # Condorcet consistency with CM
        df_condorcet_consistency_despite_cm = self.df[
            self.df['Criterion'] == 'elects_condorcet_winner_rk_even_with_cm_'
            ].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)',
            aggfunc=np.sum
        )
        df_condorcet_consistency_despite_cm = df_condorcet_consistency_despite_cm.loc[
                                              :, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]
        # Condorcet violation (sincere / with CM)
        df_violation = pd.DataFrame()
        df_violation[('Sincere', 'Rate (lower bound)')] = - df_condorcet_consistency[
            'Rate (lower bound)'] / cum_rate_condorcet_exists + 1
        df_violation[('CM', 'Rate (lower bound)')] = - df_condorcet_consistency_despite_cm[
            'Rate (upper bound)'] / cum_rate_condorcet_exists + 1
        df_violation[('CM', 'Rate (upper bound)')] = - df_condorcet_consistency_despite_cm[
            'Rate (lower bound)'] / cum_rate_condorcet_exists + 1
        df_violation[('CM', 'Rate (uncertainty)')] = df_condorcet_consistency_despite_cm[
                                                         'Rate (uncertainty)'] / cum_rate_condorcet_exists
        df_violation.columns = pd.MultiIndex.from_tuples(df_violation.columns)
        df_violation = df_violation.loc[self.rules_order, :]
        df_violation.sort_values(by=[
            ('CM', 'Rate (lower bound)'), ('CM', 'Rate (upper bound)'), ('Sincere', 'Rate (lower bound)')
        ], inplace=True)
        # Plot
        x = np.arange(len(df_violation.index))
        bar_width = 0.35
        fig, ax = plt.subplots(figsize=self.figsize)
        ax.bar(x - bar_width / 2, df_violation[('Sincere', 'Rate (lower bound)')], bar_width, label='Sincere')
        ax.bar(x + bar_width / 2, df_violation[('CM', 'Rate (lower bound)')], bar_width, label='With CM',
               yerr=(np.zeros(len(df_violation.index)), df_violation[('CM', 'Rate (uncertainty)')]))
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        ax.set_xticks(x)
        ax.set_xticklabels(df_violation.index)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('Condorcet violation rate')
        ax.legend()
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_violation

    def loss_social_welfare_bar_plot(self, tikz_file='loss_social_welfare_bar_plot.tex'):
        # Sincere social welfare
        df_social_welfare_sincere = self.df[
            self.df['Criterion'] == 'relative_social_welfare_mean'
        ].pivot_table(
            values=['Rate (lower bound)'],
            index='Rule (abbr)',
        )
        # Social welfare with CM
        df_social_welfare_with_cm = self.df[
            self.df['Criterion'] == 'worst_relative_welfare_with_cm_'
        ].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)',
        )
        # Loss of social welfare (sincere / with CM)
        df_loss_sw = pd.DataFrame()
        df_loss_sw[('Sincere', 'Rate (lower bound)')] = - df_social_welfare_sincere['Rate (lower bound)'] + 1
        df_loss_sw[('CM', 'Rate (lower bound)')] = - df_social_welfare_with_cm['Rate (upper bound)'] + 1
        df_loss_sw[('CM', 'Rate (upper bound)')] = - df_social_welfare_with_cm['Rate (lower bound)'] + 1
        df_loss_sw[('CM', 'Rate (uncertainty)')] = df_social_welfare_with_cm['Rate (uncertainty)']
        df_loss_sw.columns = pd.MultiIndex.from_tuples(df_loss_sw.columns)
        df_loss_sw = df_loss_sw.loc[self.rules_order, :]
        df_loss_sw.sort_values(by=[
            ('CM', 'Rate (lower bound)'), ('CM', 'Rate (upper bound)'), ('Sincere', 'Rate (lower bound)')
        ], inplace=True)
        # Plot
        x = np.arange(len(df_loss_sw.index))
        bar_width = 0.35
        fig, ax = plt.subplots(figsize=self.figsize)
        ax.bar(x - bar_width / 2, df_loss_sw[('Sincere', 'Rate (lower bound)')], bar_width, label='Sincere')
        ax.bar(x + bar_width / 2, df_loss_sw[('CM', 'Rate (lower bound)')], bar_width, label='Despite CM',
               yerr=(np.zeros(len(df_loss_sw.index)), df_loss_sw[('CM', 'Rate (uncertainty)')]))
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        ax.set_xticks(x)
        ax.set_xticklabels(df_loss_sw.index)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('Loss of normalized social welfare')
        ax.legend(loc='upper left')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_loss_sw

    def nb_candidates_cm_line_plot(self, rules=None, tikz_file='nb_candidates_cm_line_plot.tex'):
        # Generate the df
        df_nb_candidates_cm = self.df[
            self.df['Criterion'] == 'nb_candidates_cm_'
        ].pivot_table(
            values=['Rate (lower bound)'],
            index='Rule (abbr)',
            columns='C',
        )
        df_nb_candidates_cm = df_nb_candidates_cm.loc[self.rules_order, 'Rate (lower bound)']
        if rules is not None:
            df_nb_candidates_cm = df_nb_candidates_cm.loc[
                [rule for rule in df_nb_candidates_cm.index if rule in rules], :]
        # Sort the df
        df_nb_candidates_cm['mean'] = df_nb_candidates_cm.mean(axis=1)
        df_nb_candidates_cm.sort_values('mean', inplace=True, ascending=False)
        df_nb_candidates_cm.drop('mean', axis=1, inplace=True)
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        df_nb_candidates_cm.T.plot(ax=ax)
        plt.grid()
        ax.set_axisbelow(True)
        plt.xlabel('Number of candidates')
        plt.xlim(3, df_nb_candidates_cm.columns[-1])
        plt.yticks(range(df_nb_candidates_cm.columns[-1]))
        plt.ylabel('Average number of candidates who can benefit from CM')
        plt.ylim(0, df_nb_candidates_cm.columns[-1] - .5)
        plt.legend(loc='center right')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_nb_candidates_cm

    def nb_candidates_cm_bar_plot(self, tikz_file='nb_candidates_cm_bar_plot.tex'):
        # Raw data
        df_nb_candidates_cm = self.df[
            self.df['Criterion'] == 'nb_candidates_cm_'
        ]
        # Raw data: proportion of candidates
        df_proportion_candidates_cm = pd.DataFrame()
        df_proportion_candidates_cm['Rule (abbr)'] = df_nb_candidates_cm['Rule (abbr)']
        df_proportion_candidates_cm['Rate (lower bound)'] = df_nb_candidates_cm['Rate (lower bound)'] / (
                df_nb_candidates_cm['C'] - 1)
        df_proportion_candidates_cm['Rate (upper bound)'] = df_nb_candidates_cm['Rate (upper bound)'] / (
                df_nb_candidates_cm['C'] - 1)
        df_proportion_candidates_cm['Rate (uncertainty)'] = df_nb_candidates_cm['Rate (uncertainty)'] / (
                df_nb_candidates_cm['C'] - 1)
        # Pivot table
        df_plot = df_proportion_candidates_cm.pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)',
        )
        df_plot = df_plot.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]
        df_plot.sort_values(by=['Rate (lower bound)', 'Rate (upper bound)'], inplace=True)
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        plt.bar(
            df_plot.index,
            df_plot['Rate (lower bound)'],
            yerr=(np.zeros(len(df_plot.index)), df_plot['Rate (uncertainty)'])
        )
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('Average proportion of candidates who can benefit from CM')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_plot

    def cm_power_index_bar_plot(self, tikz_file='cm_power_index_bar_plot.tex'):
        # Create the pivot table
        df_plot = self.df[self.df['Criterion'] == 'cm_power_index_'].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)'
        )
        df_plot = df_plot.loc[self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]
        df_plot.sort_values(by=['Rate (lower bound)', 'Rate (upper bound)', ], inplace=True)
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        plt.bar(
            df_plot.index,
            df_plot['Rate (lower bound)'],
            yerr=(np.zeros(len(df_plot.index)), df_plot['Rate (uncertainty)'])
        )
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        plt.xlim(-1, self.n_rules)
        y_max = np.ceil(df_plot['Rate (lower bound)'].max()) + .5
        plt.ylim(0, y_max)
        plt.ylabel('CM power index')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_plot

    def cm_complexity_index_bar_plot(self, tikz_file='cm_complexity_index_bar_plot.tex'):
        df_tm_or_um = self.df[
            self.df['Criterion'] == 'is_tm_or_um_'
        ].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)',
            aggfunc=np.sum,
        )
        df_cm = self.df[
            self.df['Criterion'] == 'is_cm_'
        ].pivot_table(
            values=['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)'],
            index='Rule (abbr)',
            aggfunc=np.sum,
        )
        df_complexity_index = pd.DataFrame()
        df_complexity_index['Simplicity (lower bound)'] = df_tm_or_um['Rate (lower bound)'] / df_cm[
            'Rate (upper bound)']
        df_complexity_index['Simplicity (upper bound)'] = np.minimum(
            df_tm_or_um['Rate (upper bound)'] / df_cm['Rate (lower bound)'], 1)
        df_complexity_index['Rate (upper bound)'] = - df_complexity_index['Simplicity (lower bound)'] + 1
        df_complexity_index['Rate (lower bound)'] = - df_complexity_index['Simplicity (upper bound)'] + 1
        df_complexity_index['Rate (uncertainty)'] = (df_complexity_index['Rate (upper bound)']
                                                     - df_complexity_index['Rate (lower bound)'])
        df_complexity_index.drop('Simplicity (lower bound)', axis=1, inplace=True)
        df_complexity_index.drop('Simplicity (upper bound)', axis=1, inplace=True)
        df_complexity_index = df_complexity_index.loc[
            self.rules_order, ['Rate (lower bound)', 'Rate (upper bound)', 'Rate (uncertainty)']]
        df_complexity_index.sort_values(by=['Rate (lower bound)', 'Rate (upper bound)'], ascending=False, inplace=True)
        # Plot
        fig, ax = plt.subplots(figsize=self.figsize)
        plt.bar(
            df_complexity_index.index,
            df_complexity_index['Rate (lower bound)'],
            yerr=(np.zeros(len(df_complexity_index.index)), df_complexity_index['Rate (uncertainty)'])
        )
        plt.grid(axis='y')
        ax.set_axisbelow(True)
        plt.xlim(-1, self.n_rules)
        plt.yticks(np.arange(0, 1.1, step=0.1))
        plt.ylabel('CM complexity index')
        tikzplotlib.save(self.tikz_directory / (self.prefix_tikz_file + tikz_file))
        return df_complexity_index

    def df_computation_time(self):
        df_sorted = self.df.sort_values(by='Computation time', ascending=False)
        df_sorted.drop(['Rule (abbr)', 'Rule (class)', 'Candidate name', 'Culture type'], axis=1, inplace=True)
        for i in range(1, 5):
            df_sorted.drop([f'Culture parameter {i}', f'Culture parameter {i} value'], axis=1, inplace=True)
        return df_sorted

    def df_computation_time_cumulative(self):
        df_cum_time = self.df.pivot_table(
            values='Computation time',
            index=['Rule (abbr)', 'Criterion'],
            aggfunc=np.sum,
        )
        df_cum_time.reset_index(inplace=True)
        df_cum_time.sort_values(by='Computation time', ascending=False, inplace=True)
        return df_cum_time



import numpy as np
from svvamp import RuleMajorityJudgment, Profile


def test_min_grade_setter():
    """
        >>> rule = RuleMajorityJudgment()
        >>> rule.min_grade = 0
        >>> rule.min_grade
        0
        >>> rule.min_grade = -42
        >>> rule.min_grade
        -42
        >>> rule.min_grade = -np.inf
        Traceback (most recent call last):
        ValueError: Unknown value for min_grade: -inf
    """
    pass


def test_max_grade_setter():
    """
        >>> rule = RuleMajorityJudgment()
        >>> rule.max_grade = 1
        >>> rule.max_grade
        1
        >>> rule.max_grade = 42
        >>> rule.max_grade
        42
        >>> rule.max_grade = np.inf
        Traceback (most recent call last):
        ValueError: Unknown value for max_grade: inf
    """
    pass


def test_step_grade_setter():
    """
        >>> rule = RuleMajorityJudgment()
        >>> rule.step_grade = 0
        >>> rule.step_grade
        0
        >>> rule.step_grade = 0.1
        >>> rule.step_grade
        0.1
        >>> rule.step_grade = np.inf
        Traceback (most recent call last):
        ValueError: Unknown value for step_grade: inf
    """
    pass


def test_rescale_grades_setter():
    """
        >>> rule = RuleMajorityJudgment()
        >>> rule.rescale_grades = True
        >>> rule.rescale_grades
        True
        >>> rule.rescale_grades = False
        >>> rule.rescale_grades
        False
        >>> rule.rescale_grades = 'unexpected value'
        Traceback (most recent call last):
        ValueError: Unknown value for rescale_grades: unexpected value
    """
    pass


def test_no_rescale_grades():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMajorityJudgment(rescale_grades=False)(profile)
        >>> rule.w_
        0
    """
    pass


def test_step_grade_not_equal_to_zero():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMajorityJudgment(step_grade=0.1)(profile)
        >>> rule.w_
        0
    """
    pass


def test_tm_um_c():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMajorityJudgment()(profile)
        >>> rule.is_tm_c_(0)
        False
        >>> rule.is_um_c_(0)
        False
    """
    pass
